import asyncio

from trame.app import asynchronous
from trame.decorators import change
from trame.widgets import client, html
from trame.widgets import vuetify3 as v3

from e3sm_quickview.utils import js

DENSITY = {
    "adjust-layout": "compact",
    "adjust-databounds": "default",
    "select-slice-time": "default",
    "animation-controls": "compact",
}

SIZES = {
    "adjust-layout": 98,  # Two rows: 49 + 49
    "adjust-databounds": 65,
    "select-slice-time": 70,
    "animation-controls": 49,
}

VALUES = list(DENSITY.keys())

DEFAULT_STYLES = {
    "color": "white",
    "classes": "border-b-thin",
}


def to_kwargs(value):
    return {
        "v_show": js.is_active(value),
        "density": DENSITY[value],
        **DEFAULT_STYLES,
    }


class Layout(html.Div):
    def __init__(self, apply_size=None):
        style = to_kwargs("adjust-layout")
        style["style"] = "background: rgb(var(--v-theme-surface));"
        super().__init__(**style)

        with self:
            # First row - existing layout controls
            with v3.VToolbar(density="compact", color="white", classes="border-b-thin"):
                v3.VIcon("mdi-collage", classes="px-6 opacity-50")
                v3.VLabel("Viewport layout", classes="text-subtitle-2")
                v3.VSpacer()

                v3.VSlider(
                    v_model=("aspect_ratio", 2),
                    prepend_icon="mdi-aspect-ratio",
                    min=1,
                    max=2,
                    step=0.1,
                    density="compact",
                    hide_details=True,
                    style="max-width: 400px;",
                )
                v3.VSpacer()

                # ------------------------------------------------------------
                # Add tooltip for keyboard shortcut??
                # ------------------------------------------------------------
                # with v3.VTooltip(location="bottom"):
                #    with v3.Template(v_slot_activator="{ props }"):
                v3.VHotkey(keys="g", variant="contained", classes="mr-1")
                v3.VCheckbox(
                    # v_bind="props",
                    v_model=("layout_grouped", True),
                    label=("layout_grouped ? 'Grouped' : 'Uniform'",),
                    hide_details=True,
                    inset=True,
                    false_icon="mdi-apps",
                    true_icon="mdi-focus-field",
                    density="compact",
                )
                # with html.Span("Keyboard shortcut"):
                #     v3.VHotkey(theme="dark", keys="g", variant="contained", inline=True, classes="ml-2 mt-n2")
                # ------------------------------------------------------------

                with v3.VBtn(
                    "Size",
                    classes="text-none mx-4",
                    prepend_icon="mdi-view-module",
                    append_icon="mdi-menu-down",
                ):
                    with v3.VMenu(activator="parent"):
                        with v3.VList(density="compact"):
                            with v3.VListItem(
                                title="Auto flow",
                                click=(
                                    apply_size,
                                    "['flow']",
                                ),
                            ):
                                with v3.Template(v_slot_append=True):
                                    v3.VHotkey(
                                        keys="=",
                                        variant="contained",
                                        inline=True,
                                        classes="ml-6 mt-n1",
                                    )
                            with v3.VListItem(
                                title="Auto",
                                click=(
                                    apply_size,
                                    "[0]",
                                ),
                            ):
                                with v3.Template(v_slot_append=True):
                                    v3.VHotkey(
                                        keys="0",
                                        variant="contained",
                                        inline=True,
                                        classes="ml-6 mt-n1",
                                    )
                            with v3.VListItem(
                                title="Full Width",
                                click=(
                                    apply_size,
                                    "[1]",
                                ),
                            ):
                                with v3.Template(v_slot_append=True):
                                    v3.VHotkey(
                                        keys="1",
                                        variant="contained",
                                        inline=True,
                                        classes="ml-6 mt-n1",
                                    )
                            with v3.VListItem(
                                title="2 Columns",
                                click=(
                                    apply_size,
                                    "[2]",
                                ),
                            ):
                                with v3.Template(v_slot_append=True):
                                    v3.VHotkey(
                                        keys="2",
                                        variant="contained",
                                        inline=True,
                                        classes="ml-6 mt-n1",
                                    )
                            with v3.VListItem(
                                title="3 Columns",
                                click=(
                                    apply_size,
                                    "[3]",
                                ),
                            ):
                                with v3.Template(v_slot_append=True):
                                    v3.VHotkey(
                                        keys="3",
                                        variant="contained",
                                        inline=True,
                                        classes="ml-6 mt-n1",
                                    )
                            with v3.VListItem(
                                title="4 Columns",
                                click=(
                                    apply_size,
                                    "[4]",
                                ),
                            ):
                                with v3.Template(v_slot_append=True):
                                    v3.VHotkey(
                                        keys="4",
                                        variant="contained",
                                        inline=True,
                                        classes="ml-6 mt-n1",
                                    )
                            with v3.VListItem(
                                title="6 Columns",
                                click=(
                                    apply_size,
                                    "[6]",
                                ),
                            ):
                                with v3.Template(v_slot_append=True):
                                    v3.VHotkey(
                                        keys="6",
                                        variant="contained",
                                        inline=True,
                                        classes="ml-6 mt-n1",
                                    )

            # Second row - column toggle buttons
            with v3.VToolbar(density="compact", color="white", classes="border-b-thin"):
                v3.VLabel("Columns:", classes="text-subtitle-2 px-4")
                for comp_type in ["ctrl", "test", "diff", "comp1", "comp2"]:
                    v3.VBtn(
                        comp_type.upper(),
                        size="small",
                        variant=("selected_columns.includes('{0}') ? 'flat' : 'outlined'".format(comp_type),),
                        color=("selected_columns.includes('{0}') ? 'primary' : 'default'".format(comp_type),),
                        classes="mx-1",
                        click=(
                            f"selected_columns.includes('{comp_type}') ? "
                            f"selected_columns = selected_columns.filter(c => c !== '{comp_type}') : "
                            f"selected_columns = [...selected_columns, '{comp_type}']"
                        ),
                    )
                v3.VSpacer()
                v3.VBtn(
                    "Show All",
                    size="small",
                    variant="text",
                    click="selected_columns = ['ctrl', 'test', 'diff', 'comp1', 'comp2']",
                )


class Cropping(v3.VToolbar):
    def __init__(self):
        super().__init__(**to_kwargs("adjust-databounds"))

        with self:
            v3.VIcon("mdi-web", classes="pl-6 opacity-50")
            with v3.VRow(classes="ma-0 px-2 align-center"):
                with v3.VCol(cols=6):
                    with v3.VRow(classes="mx-2 my-0"):
                        v3.VLabel(
                            "Longitude",
                            classes="text-subtitle-2",
                        )
                        v3.VSpacer()
                        v3.VLabel(
                            "{{ crop_longitude }}",
                            classes="text-body-2",
                        )
                    v3.VRangeSlider(
                        v_model=("crop_longitude", [-180, 180]),
                        min=-180,
                        max=180,
                        step=1,
                        density="compact",
                        hide_details=True,
                    )
                with v3.VCol(cols=6):
                    with v3.VRow(classes="mx-2 my-0"):
                        v3.VLabel(
                            "Latitude",
                            classes="text-subtitle-2",
                        )
                        v3.VSpacer()
                        v3.VLabel(
                            "{{ crop_latitude }}",
                            classes="text-body-2",
                        )
                    v3.VRangeSlider(
                        v_model=("crop_latitude", [-90, 90]),
                        min=-90,
                        max=90,
                        step=1,
                        density="compact",
                        hide_details=True,
                    )


class DataSelection(html.Div):
    def __init__(self):
        style = to_kwargs("select-slice-time")
        # Use style instead of d-flex class to avoid !important override of v-show
        # Add background color to match VToolbar appearance
        style["style"] = (
            "display: flex; align-items: center; background: rgb(var(--v-theme-surface));"
        )
        super().__init__(**style)

        with self:
            v3.VIcon("mdi-tune-variant", classes="ml-3 mr-2 opacity-50")

            with v3.VRow(classes="ma-0 pr-2 flex-wrap flex-grow-1", dense=True):
                # Debug: Show animation_tracks array
                # html.Div("Animation Tracks: {{ JSON.stringify(animation_tracks) }}", classes="col-12")
                # Each track gets a column (3 per row)
                with v3.VCol(
                    cols=4,
                    v_for="(track, idx) in animation_tracks",
                    key="idx",
                    classes="pa-2",
                ):
                    with client.Getter(name=("track.value",), value_name="t_values"):
                        with client.Getter(
                            name=("track.value + '_idx'",), value_name="t_idx"
                        ):
                            with v3.VRow(classes="ma-0 align-center", dense=True):
                                v3.VLabel(
                                    "{{track.title}}",
                                    classes="text-subtitle-2",
                                )
                                v3.VSpacer()
                                v3.VLabel(
                                    "{{ parseFloat(t_values[t_idx]).toFixed(2) }} hPa (k={{ t_idx }})",
                                    classes="text-body-2",
                                )
                            v3.VSlider(
                                model_value=("t_idx",),
                                update_modelValue=(
                                    self.on_update_slider,
                                    "[track.value, $event]",
                                ),
                                min=0,
                                # max=100,#("get(track.value).length - 1",),
                                max=("t_values.length - 1",),
                                step=1,
                                density="compact",
                                hide_details=True,
                            )

    def on_update_slider(self, dimension, index, *_, **__):
        with self.state:
            self.state[f"{dimension}_idx"] = index


class Animation(v3.VToolbar):
    def __init__(self):
        super().__init__(**to_kwargs("animation-controls"))

        with self:
            v3.VIcon(
                "mdi-video",
                classes="px-6 opacity-50",
            )
            with v3.VRow(classes="ma-0 px-2 align-center"):
                v3.VSelect(
                    v_model=("animation_track", "timestamps"),
                    items=("animation_tracks", []),
                    flat=True,
                    variant="plain",
                    hide_details=True,
                    density="compact",
                    style="max-width: 10rem;",
                )
                v3.VDivider(vertical=True, classes="mx-2")
                v3.VSlider(
                    v_model=("animation_step", 1),
                    min=0,
                    max=("amimation_step_max", 0),
                    step=1,
                    hide_details=True,
                    density="compact",
                    classes="mx-4",
                )
                v3.VDivider(vertical=True, classes="mx-2")
                v3.VIconBtn(
                    icon="mdi-page-first",
                    flat=True,
                    disabled=("animation_step === 0",),
                    click="animation_step = 0",
                )
                v3.VIconBtn(
                    icon="mdi-chevron-left",
                    flat=True,
                    disabled=("animation_step === 0",),
                    click="animation_step = Math.max(0, animation_step - 1)",
                )
                v3.VIconBtn(
                    icon="mdi-chevron-right",
                    flat=True,
                    disabled=("animation_step === amimation_step_max",),
                    click="animation_step = Math.min(amimation_step_max, animation_step + 1)",
                )
                v3.VIconBtn(
                    icon="mdi-page-last",
                    disabled=("animation_step === amimation_step_max",),
                    flat=True,
                    click="animation_step = amimation_step_max",
                )
                v3.VDivider(vertical=True, classes="mx-2")
                v3.VIconBtn(
                    icon=("animation_play ? 'mdi-stop' : 'mdi-play'",),
                    flat=True,
                    click="animation_play = !animation_play",
                )

    @change("animation_track")
    def _on_animation_track_change(self, animation_track, **_):
        self.state.animation_step = 0
        self.state.amimation_step_max = 0

        if animation_track:
            self.state.amimation_step_max = len(self.state[animation_track]) - 1

    @change("animation_step")
    def _on_animation_step(self, animation_track, animation_step, **_):
        if animation_track:
            self.state[f"{animation_track}_idx"] = animation_step

    @change("animation_play")
    def _on_animation_play(self, animation_play, **_):
        if animation_play:
            asynchronous.create_task(self._run_animation())

    async def _run_animation(self):
        with self.state as s:
            while s.animation_play:
                await asyncio.sleep(0.1)
                if s.animation_step < s.amimation_step_max:
                    with s:
                        s.animation_step += 1
                    await self.server.network_completion
                else:
                    s.animation_play = False



"""Server proxy configuration for CompareView in JupyterLab."""

import os


def setup_compareview():
    """Configure jupyter-server-proxy for CompareView."""
    icon_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "icons",
        "compareview.png",
    )

    return {
        "command": [
            "compareview",
            "--server",
            "--port",
            "{port}",
            "--host",
            "127.0.0.1",
        ],
        "timeout": 30,
        "launcher_entry": {
            "enabled": True,
            "title": "CompareView",
            "icon_path": icon_path,
            "category": "Other",
        },
        "new_browser_tab": False,
    }



"""CompareView: E3SM CompareView application."""

# TODO(decouple): Revisit "quickview" naming/usage in Tauri configs after refactor.
__version__ = "1.2.1"
__author__ = "Kitware Inc."
__license__ = "Apache-2.0"



"""Jupyter integration for CompareView via jupyter-server-proxy."""

from .proxy import setup_compareview

# Backward compatibility alias.
setup_quickview = setup_compareview

__all__ = ["setup_compareview", "setup_quickview"]



import json
import re
from pathlib import Path
from paraview import simple
from trame.widgets import vuetify3 as v3, html
from trame.app import TrameComponent

DIRECTORY = dict(icon="mdi-folder", type="directory")
GROUP = dict(icon="mdi-file-document-multiple-outline", type="group")
FILE = dict(icon="mdi-file-document-outline", type="file")

HEADERS = [
    {"title": "Name", "align": "start", "key": "name", "sortable": False},
    {"title": "Size", "align": "end", "key": "size", "sortable": False},
    {"title": "Date", "align": "end", "key": "modified", "sortable": False},
]


def sort_by_name(e):
    return e.get("name")


def to_type(e):
    return e.get("type", "")


def to_suffix(e):
    return Path(e.get("name", "")).suffix


class ParaViewFileBrowser(TrameComponent):
    def __init__(
        self,
        server,
        prefix="pv_files",
        home=None,
        current=None,
        exclude=r"^\.|~$|^\$",
        group=r"[0-9]+\.",
    ):
        super().__init__(server)
        self._prefix = prefix

        self._enable_groups = True
        self._home_path = Path(home).resolve() if home else Path.home()
        self._current_path = Path(current).resolve() if current else self._home_path
        self.pattern_exclude = re.compile(exclude)
        self.pattern_group = re.compile(group)

        # Disable state import by default
        self.set("is_state_file", False)

        # Initialize simulation files list (instead of single file)
        self.set("data_simulation_files", [])

        self._pxm = simple.servermanager.ProxyManager()
        self._proxy_listing = self._pxm.NewProxy("misc", "ListDirectory")
        self._proxy_directories = simple.servermanager.VectorProperty(
            self._proxy_listing, self._proxy_listing.GetProperty("DirectoryList")
        )
        self._proxy_files = simple.servermanager.VectorProperty(
            self._proxy_listing, self._proxy_listing.GetProperty("FileList")
        )

        # Initialize trame state
        self.update_listing()

    def name(self, name):
        return f"{self._prefix}_{name}"

    def set(self, name, value):
        self.state[self.name(name)] = value

    def get(self, name):
        return self.state[self.name(name)]

    def update_listing(self, selection=None):
        with self.state:
            self.set("active", -1)
            self.set("listing", self.listing)
            self.set("selected", selection)

    @property
    def enable_groups(self):
        return self._enable_groups

    @enable_groups.setter
    def enable_groups(self, v):
        self._enable_groups = v

    @property
    def listing(self):
        directories = []
        files = []
        groups = []
        g_map = {}

        self._proxy_listing.List(str(self._current_path.resolve()))
        self._proxy_listing.UpdatePropertyInformation()

        # Files + Groups
        file_listing = []
        if len(self._proxy_files) > 1:
            file_listing = self._proxy_files.GetData()
        if len(self._proxy_files) == 1:
            file_listing.append(self._proxy_files.GetData())
        file_listing = [
            file_name
            for file_name in file_listing
            if not re.search(self.pattern_exclude, file_name)
        ]
        for file_name in file_listing:
            f = self._current_path / file_name
            stats = f.stat()

            # Group or file?
            file_split = re.split(self.pattern_group, file_name)
            if self.enable_groups and len(file_split) == 2:
                # Group
                g_name = "*.".join(file_split)
                if g_name not in g_map:
                    g_entry = dict(
                        name=g_name,
                        modified=stats.st_mtime,
                        size=0,
                        files=[],
                        **GROUP,
                    )
                    g_map[g_name] = g_entry
                    groups.append(g_entry)

                g_map[g_name]["size"] += stats.st_size
                g_map[g_name]["files"].append(file_name)
                # Many need to sort files???
            else:
                # File
                files.append(
                    dict(
                        name=f.name,
                        modified=stats.st_mtime,
                        size=stats.st_size,
                        **FILE,
                    )
                )

        # Directories
        dir_listing = []
        if len(self._proxy_directories) > 1:
            dir_listing = self._proxy_directories.GetData()
        if len(self._proxy_directories) == 1:
            dir_listing.append(self._proxy_directories.GetData())
        dir_listing = [
            dir_name
            for dir_name in dir_listing
            if not re.search(self.pattern_exclude, dir_name)
        ]
        for dir_name in dir_listing:
            f = self._current_path / dir_name
            directories.append(
                dict(name=f.name, modified=f.stat().st_mtime, **DIRECTORY)
            )

        # Sort content
        directories.sort(key=sort_by_name)
        groups.sort(key=sort_by_name)
        files.sort(key=sort_by_name)

        return [
            {**e, "index": i} for i, e in enumerate([*directories, *groups, *files])
        ]

    def open_entry(self, entry):
        entry_type = entry.get("type")
        if entry_type == "directory":
            self._current_path = self._current_path / entry.get("name")
            self.update_listing()
            return entry_type, str(self._current_path)
        if entry_type == "group":
            files = entry.get("files", [])
            self.update_listing()
            return entry, [str(self._current_path / f) for f in files]
        if entry_type == "file":
            file = self._current_path / entry.get("name")
            file_name = file.name.lower()
            full_path = str(file)
            if "connectivity_" in file_name:
                self.set("data_connectivity", full_path)
            else:
                # Add to simulation files list
                current_files = self.get("data_simulation_files") or []
                if full_path not in current_files:
                    # Create a new list to trigger state change
                    self.set("data_simulation_files", [*current_files, full_path])
                # Also set the legacy single file variable for backward compatibility
                self.set("data_simulation", full_path)
            self.update_listing(full_path)
            return entry_type, full_path

        return None

    @property
    def active_path(self):
        entry = self.get("listing")[self.get("active")]
        return str(self._current_path / entry.get("name"))

    def set_data_connectivity(self, value=None):
        self.set("data_connectivity", value or self.active_path)

    # TODO: remove this method (used by legacy CLI)
    def set_data_simulation(self, value=None):
        """Legacy method for backward compatibility - adds to simulation files list"""
        file_path = value or self.active_path
        self.add_simulation_file(file_path)

    def add_simulation_file(self, file_path=None):
        file_path = file_path or self.active_path
        current_files = self.get("data_simulation_files") or []
        if file_path not in current_files:
            # Create a new list to trigger state change
            self.set("data_simulation_files", [*current_files, file_path])
        # Set the single file variable to the most recently added file
        self.set("data_simulation", file_path)

    def remove_simulation_file(self, file_path):
        current_files = self.get("data_simulation_files") or []
        if file_path in current_files:
            # Create a new list to trigger state change
            new_files = [f for f in current_files if f != file_path]
            self.set("data_simulation_files", new_files)
            # Update the single file variable
            if new_files:
                self.set("data_simulation", new_files[-1])
            else:
                self.set("data_simulation", "")

    def clear_simulation_files(self):
        self.set("data_simulation_files", [])
        self.set("data_simulation", "")

    def goto_home(self):
        self._current_path = self._home_path
        self.update_listing()

    def goto_parent(self):
        self._current_path = self._current_path.parent
        self.update_listing()

    def open_dataset(self, entry):
        event = {}
        if to_type(entry) == "group":
            files = [str(self._current_path / f) for f in entry.get("files")]
            source = simple.OpenDataFile(files)
            representation = simple.Show(source)
            view = simple.Render()
            event = dict(
                source=source, representation=representation, view=view, type="group"
            )
        else:
            source = simple.OpenDataFile(str(self._current_path / entry.get("name")))
            representation = simple.Show(source)
            view = simple.Render()
            event = dict(
                source=source, representation=representation, view=view, type="dataset"
            )

        return event

    def select_entry(self, entry):
        with self.state as state:
            state[f"{self._prefix}_active"] = entry.get("index", 0) if entry else -1
            file_path = Path(self.active_path)

            # Check if it is a state file
            if file_path.suffix == ".json" and file_path.exists():
                state_content = json.loads(file_path.read_text())
                self.set(
                    "is_state_file",
                    all(
                        (
                            k in state_content
                            for k in [
                                "files",
                                "variables-selection",
                                "layout",
                                "data-selection",
                                "views",
                            ]
                        )
                    ),
                )
            else:
                self.set("is_state_file", False)

    def load_data_files(self, **_):
        self.set("loading", True)
        print("Load files:")
        simulation_files = self.get("data_simulation_files") or []
        print(" - simulation files:", simulation_files)
        print(" - connectivity:", self.get("data_connectivity"))
        self.ctrl.file_selection_load(
            simulation_files, self.get("data_connectivity")
        )

    def import_state_file(self):
        self.set("state_loading", True)

        state_content = json.loads(Path(self.active_path).read_text())
        self.ctrl.import_state(state_content)

    def cancel(self):
        self.ctrl.file_selection_cancel()

    def loading_completed(self, valid):
        with self.state:
            self.set("loading", False)
            self.set("error", not valid)

    def ui(self):
        with v3.VCard(rounded="lg"):
            with v3.VCardTitle("File loading", classes="d-flex align-center px-3"):
                v3.VSpacer()
                v3.VBtn(
                    icon="mdi-home",
                    variant="flat",
                    size="small",
                    click=self.goto_home,
                )
                v3.VBtn(
                    icon="mdi-folder-upload-outline",
                    variant="flat",
                    size="small",
                    click=self.goto_parent,
                )
                v3.VTextField(
                    v_model=self.name("filter"),
                    hide_details=True,
                    color="primary",
                    placeholder="filter",
                    density="compact",
                    variant="outlined",
                    classes="ml-2",
                    prepend_inner_icon="mdi-magnify",
                    clearable=True,
                )

            with v3.VCardText(
                classes="rounded-lg border border-opacity-25 pa-0 mx-3 my-0 overflow-hidden"
            ):
                style_align_center = "d-flex align-center "
                with v3.VDataTable(
                    density="compact",
                    fixed_header=True,
                    headers=(self.name("headers"), HEADERS),
                    items=(self.name("listing"), []),
                    height="calc(80vh - 20rem)",
                    style="user-select: none; cursor: pointer;",
                    hover=True,
                    search=(self.name("filter"), ""),
                    items_per_page=-1,
                ):
                    v3.Template(raw_attrs=["v-slot:bottom"])
                    with v3.Template(raw_attrs=['v-slot:item="{ index, item }"']):
                        with v3.VDataTableRow(
                            index=("index",),
                            item=("item",),
                            click=(self.select_entry, "[item]"),
                            dblclick=(self.open_entry, "[item]"),
                            classes=(
                                f"{{ 'bg-grey': item.index === {self.name('active')}, 'cursor-pointer': 1 }}",
                            ),
                        ):
                            with v3.Template(raw_attrs=["v-slot:item.name"]):
                                with html.Div(classes=style_align_center):
                                    v3.VIcon(
                                        "{{ item.icon }}",
                                        size="small",
                                        classes="mr-2",
                                    )
                                    html.Div("{{ item.name }}")

                            with v3.Template(raw_attrs=["v-slot:item.size"]):
                                with html.Div(
                                    classes=style_align_center + " justify-end",
                                ):
                                    html.Div(
                                        "{{ utils.fmt.bytes(item.size, 0) }}",
                                        v_if="item.size",
                                    )
                                    html.Div(" - ", v_else=True)

                            with v3.Template(raw_attrs=["v-slot:item.modified"]):
                                with html.Div(
                                    classes=style_align_center + " justify-end",
                                ):
                                    html.Div(
                                        "{{ new Date(item.modified * 1000).toDateString() }}"
                                    )

            with v3.VCol():
                html.Label(
                    "Simulation Files",
                    classes="text-subtitle-1 font-weight-medium d-block",
                )
                # Display list of simulation files with remove buttons
                with html.Div(
                    v_if=f"{self.name('data_simulation_files')}.length > 0",
                    classes="mb-2",
                ):
                    with html.Div(
                        v_for=f"file in {self.name('data_simulation_files')}",
                        key="file",
                        style="display: inline-block;",
                    ):
                        with v3.VChip(
                            closable=True,
                            click_close=(self.remove_simulation_file, "[file]"),
                            classes="ma-1",
                            size="small",
                        ):
                            html.Span("{{ file.split('/').pop() }}")

                html.Div(
                    "No simulation files selected",
                    v_if=f"{self.name('data_simulation_files')}.length === 0",
                    classes="text-caption text-disabled mb-2",
                )
                html.Label(
                    "Connectivity File",
                    classes="text-subtitle-1 font-weight-medium d-block",
                )
                v3.VTextField(
                    v_model=(self.name("data_connectivity"), ""),
                    density="compact",
                    variant="outlined",
                    disabled=True,
                    messages="The horizontal grids used by EAM are cubed spheres. Since these are unstructured grids, CompareView needs to know how to map data to the globe. Therefore, for each simulation data file, a connectivity file needs to be provided.",
                )

            v3.VDivider()
            with v3.VCardActions(classes="pa-3"):
                v3.VBtn(
                    classes="text-none",
                    variant="tonal",
                    text="Add Simulation",
                    prepend_icon="mdi-database-plus",
                    disabled=(
                        f"{self.name('listing')}[{self.name('active')}]?.type !== 'file'",
                    ),
                    click=self.add_simulation_file,
                )
                v3.VBtn(
                    classes="text-none",
                    text="Connectivity",
                    variant="tonal",
                    prepend_icon="mdi-vector-polyline-plus",
                    disabled=(
                        f"{self.name('listing')}[{self.name('active')}]?.type !== 'file'",
                    ),
                    click=self.set_data_connectivity,
                )
                v3.VBtn(
                    classes="text-none",
                    text="Reset",
                    variant="tonal",
                    prepend_icon="mdi-close-octagon-outline",
                    click=f"{self.name('data_connectivity')}='';{self.name('data_simulation')}='';{self.name('data_simulation_files')}=[];{self.name('error')}=false",
                )
                v3.VSpacer()
                v3.VBtn(
                    border=True,
                    classes="text-none",
                    color="surface",
                    text="Cancel",
                    variant="flat",
                    click=self.cancel,
                )
                v3.VBtn(
                    disabled=(f"!{self.name('is_state_file')}",),
                    loading=(self.name("state_loading"), False),
                    classes="text-none",
                    color="primary",
                    text="Import state file",
                    variant="flat",
                    click=self.import_state_file,
                )
                v3.VBtn(
                    classes="text-none",
                    color=(f"{self.name('error')} ? 'error' : 'primary'",),
                    text="Load files",
                    variant="flat",
                    disabled=(
                        f"{self.name('data_simulation_files')}.length === 0 || !{self.name('data_connectivity')} || {self.name('error')}",
                    ),
                    loading=(self.name("loading"), False),
                    click=self.load_data_files,
                )



import fnmatch
import json
import os

import e3sm_quickview


from paraview.simple import (
    FindSource,
    LoadPlugin,
    OutputPort,
    Contour,
    LegacyVTKReader,
)

from vtkmodules.vtkCommonCore import vtkLogger

from collections import defaultdict


# Define a VTK error observer
class ErrorObserver:
    def __init__(self):
        self.error_occurred = False
        self.error_message = ""

    def __call__(self, obj, event):
        self.error_occurred = True

    def clear(self):
        self.error_occurred = False


class EAMVisSource:
    def __init__(self):
        # flag to check if the pipeline is valid
        # this is set to true when the pipeline is updated
        # and the data is available
        self.valid = False

        self.ctrl_file = None
        self.test_file = None
        self.conn_file = None

        # List of all available variables
        self.varmeta = None
        self.dimmeta = None
        self.slicing = defaultdict(int)

        self.ctrl_data = None
        self.test_data = None
        self.globe = None
        self.projection = "Cyl. Equidistant"
        self.timestamps = []
        self.center = 0.0

        self.prog_filter = None
        self.atmos_extract = None
        self.atmos_proj = None
        self.cont_extract = None
        self.cont_proj = None
        self.grid_gen = None
        self.grid_proj = None

        self.extents = [-180.0, 180.0, -90.0, 90.0]
        self.moveextents = [-180.0, 180.0, -90.0, 90.0]

        self.views = {}
        self.vars = {"surface": [], "midpoint": [], "interface": []}

        self.observer = ErrorObserver()
        try:
            plugin_dir = os.path.join(
                os.path.dirname(e3sm_quickview.__file__),
                "plugins",
            )
            plugins = fnmatch.filter(os.listdir(path=plugin_dir), "*.py")
            for plugin in plugins:
                print("Loading plugin : ", plugin)
                plugpath = os.path.abspath(os.path.join(plugin_dir, plugin))
                if os.path.isfile(plugpath):
                    LoadPlugin(plugpath, ns=globals())

            vtkLogger.SetStderrVerbosity(vtkLogger.VERBOSITY_OFF)
        except Exception as e:
            print("Error loading plugin :", e)

    def ApplyClipping(self, cliplong, cliplat):
        if not self.valid:
            return

        atmos_extract = self.atmos_extract or FindSource("AtmosExtract")
        atmos_extract.LongitudeRange = cliplong
        atmos_extract.LatitudeRange = cliplat

        cont_extract = self.cont_extract or FindSource("ContExtract")
        cont_extract.LongitudeRange = cliplong
        cont_extract.LatitudeRange = cliplat

    def UpdateCenter(self, center):
        """
        if self.center != int(center):
            self.center = int(center)

            meridian = FindSource("CenterMeridian")
            meridian.CenterMeridian = self.center

            gmeridian = FindSource("GCMeridian")
            gmeridian.CenterMeridian = self.center
        """
        pass

    def UpdateProjection(self, proj):
        if not self.valid:
            return

        atmos_proj = self.atmos_proj or FindSource("AtmosProj")
        cont_proj = self.cont_proj or FindSource("ContProj")
        grid_proj = self.grid_proj or FindSource("GridProj")
        if self.projection != proj:
            self.projection = proj
            atmos_proj.Projection = proj
            cont_proj.Projection = proj
            grid_proj.Projection = proj

    def UpdateTimeStep(self, t_index):
        if not self.valid:
            return

    def UpdatePipeline(self, time=0.0):
        if not self.valid:
            return

        atmos_proj = self.atmos_proj or FindSource("AtmosProj")
        if atmos_proj:
            atmos_proj.UpdatePipeline(time)
        self.moveextents = atmos_proj.GetDataInformation().GetBounds()

        cont_proj = self.cont_proj or FindSource("ContProj")
        if cont_proj:
            cont_proj.UpdatePipeline(time)

        atmos_extract = self.atmos_extract or FindSource("AtmosExtract")
        bounds = atmos_extract.GetDataInformation().GetBounds()

        grid_gen = self.grid_gen or FindSource("GridGen")
        if grid_gen:
            grid_gen.LongitudeRange = [bounds[0], bounds[1]]
            grid_gen.LatitudeRange = [bounds[2], bounds[3]]
        grid_proj = self.grid_proj or FindSource("GridProj")
        if grid_proj:
            grid_proj.UpdatePipeline(time)

        self.views["atmosphere_data"] = OutputPort(atmos_proj, 0)
        self.views["continents"] = OutputPort(cont_proj, 0)
        self.views["grid_lines"] = OutputPort(grid_proj, 0)

    def UpdateSlicing(self, dimension, slice):
        if self.slicing.get(dimension) == slice:
            return
        else:
            self.slicing[dimension] = slice
            if self.ctrl_data is not None and self.test_data is not None:
                x = json.dumps(self.slicing)
                self.ctrl_data.Slicing = x
                self.test_data.Slicing = x

    def Update(self, ctrl_file, test_file, conn_file, variables=[], force_reload=False):
        # Check if we need to reload
        if (
            not force_reload
            and self.ctrl_file == ctrl_file
            and self.test_file == test_file
            and self.conn_file == conn_file
        ):
            return self.valid

        # Store the file paths
        self.ctrl_file = ctrl_file
        self.test_file = test_file
        self.conn_file = conn_file

        if self.ctrl_data is None or self.test_data is None:
            ctrl_data = EAMSliceDataReader(
                registrationName="AtmosReader",
                ConnectivityFile=self.conn_file,
                DataFile=self.ctrl_file,
            )
            self.ctrl_data = ctrl_data
            vtk_obj = ctrl_data.GetClientSideObject()
            vtk_obj.AddObserver("ErrorEvent", self.observer)
            vtk_obj.GetExecutive().AddObserver("ErrorEvent", self.observer)
            ctrl_varmeta = vtk_obj.GetVariables()
            # self.observer.clear()

            test_data = EAMSliceDataReader(
                registrationName="AtmosReader2",
                ConnectivityFile=self.conn_file,
                DataFile=self.test_file,
            )
            self.test_data = test_data
            vtk_obj = test_data.GetClientSideObject()
            vtk_obj.AddObserver("ErrorEvent", self.observer)
            vtk_obj.GetExecutive().AddObserver("ErrorEvent", self.observer)
            test_varmeta = vtk_obj.GetVariables()
            self.dimmeta = vtk_obj.GetDimensions()
            self.varmeta = {
                key: val
                for key, val in ctrl_varmeta.items()
                if key in test_varmeta
            }


            for dim in self.dimmeta.keys():
                self.slicing[dim] = 0

            self.observer.clear()

        else:
            self.ctrl_data.DataFile = self.ctrl_file
            self.ctrl_data.ConnectivityFile = self.conn_file
            # self.observer.clear()

            self.test_data.DataFile = self.test_file
            self.test_data.ConnectivityFile = self.conn_file
            self.observer.clear()


        try:
            # Update pipeline and force view refresh
            self.ctrl_data.UpdatePipeline(time=0.0)
            self.test_data.UpdatePipeline(time=0.0)
            if self.observer.error_occurred:
                raise RuntimeError(
                    "Error occurred in UpdatePipeline. "
                    "Please check if the data and connectivity files exist "
                    "and are compatible"
                )

            # Ensure TimestepValues is always a list
            timestep_values = self.ctrl_data.TimestepValues
            if isinstance(timestep_values, (list, tuple)):
                self.timestamps = list(timestep_values)
            elif hasattr(timestep_values, "__iter__") and not isinstance(
                timestep_values, str
            ):
                # Handle numpy arrays or other iterables
                self.timestamps = list(timestep_values)
            else:
                # Single value - wrap in a list
                self.timestamps = (
                    [timestep_values] if timestep_values is not None else []
                )

            print("Data pre-script", self.ctrl_data, self.test_data)

            script = f"vars = {variables}\n"
            script += """
for var in vars:
    ctrl = inputs[0].CellData[f"{var}"]
    test = inputs[1].CellData[f"{var}"]

    diff = test - ctrl
    comp1 = diff / ctrl
    comp2 = (2 * diff) / (test + ctrl)

    output.CellData.append(ctrl, f'{var}')
    output.CellData.append(ctrl, f'{var}_ctrl')
    output.CellData.append(test, f'{var}_test')
    output.CellData.append(diff, f'{var}_diff')
    output.CellData.append(comp1, f'{var}_comp1')
    output.CellData.append(comp2, f'{var}_comp2')

output.CellData.append(inputs[0].CellData["area"], 'area') # needed for utils.compute.extract_avgs
"""
            print("ProgrammableFilter script:\n", script, end='')
            self.prog_filter = ProgrammableFilter(registrationName='ProgrammableFilter', Input=[self.ctrl_data, self.test_data])
            self.prog_filter.Script = script
            self.prog_filter.RequestInformationScript = ''
            self.prog_filter.RequestUpdateExtentScript = ''
            self.prog_filter.PythonPath = ''


            # Step 1: Extract and transform atmospheric data
            self.atmos_extract = EAMTransformAndExtract(  # noqa: F821
                registrationName="AtmosExtract", Input=self.prog_filter
            )
            self.atmos_extract.LongitudeRange = [-180.0, 180.0]
            self.atmos_extract.LatitudeRange = [-90.0, 90.0]
            self.atmos_extract.UpdatePipeline()
            self.extents = self.atmos_extract.GetDataInformation().GetBounds()

            # Step 2: Apply map projection to atmospheric data
            self.atmos_proj = EAMProject(  # noqa: F821
                registrationName="AtmosProj", Input=OutputPort(self.atmos_extract, 0)
            )
            self.atmos_proj.Projection = self.projection
            self.atmos_proj.Translate = 0
            self.atmos_proj.UpdatePipeline()
            self.moveextents = self.atmos_proj.GetDataInformation().GetBounds()

            # Step 3: Load and process continent outlines
            if self.globe is None:
                globe_file = os.path.join(
                    os.path.dirname(__file__), "data", "globe.vtk"
                )
                globe_reader = LegacyVTKReader(
                    registrationName="ContReader", FileNames=[globe_file]
                )
                cont_contour = Contour(
                    registrationName="ContContour", Input=globe_reader
                )
                cont_contour.ContourBy = ["POINTS", "cstar"]
                cont_contour.Isosurfaces = [0.5]
                cont_contour.PointMergeMethod = "Uniform Binning"
                self.globe = cont_contour

            # Step 4: Extract and transform continent data
            self.cont_extract = EAMTransformAndExtract(  # noqa: F821
                registrationName="ContExtract", Input=self.globe
            )
            self.cont_extract.LongitudeRange = [-180.0, 180.0]
            self.cont_extract.LatitudeRange = [-90.0, 90.0]
            # Step 5: Apply map projection to continents
            self.cont_proj = EAMProject(  # noqa: F821
                registrationName="ContProj", Input=OutputPort(self.cont_extract, 0)
            )
            self.cont_proj.Projection = self.projection
            self.cont_proj.Translate = 0
            self.cont_proj.UpdatePipeline()

            # Step 6: Generate lat/lon grid lines
            self.grid_gen = EAMGridLines(registrationName="GridGen")  # noqa: F821
            self.grid_gen.UpdatePipeline()

            # Step 7: Apply map projection to grid lines
            self.grid_proj = EAMProject(  # noqa: F821
                registrationName="GridProj", Input=OutputPort(self.grid_gen, 0)
            )
            self.grid_proj.Projection = self.projection
            self.grid_proj.Translate = 0
            self.grid_proj.UpdatePipeline()

            # Step 8: Cache all projected views for rendering
            self.views["atmosphere_data"] = OutputPort(self.atmos_proj, 0)
            self.views["continents"] = OutputPort(self.cont_proj, 0)
            self.views["grid_lines"] = OutputPort(self.grid_proj, 0)

            self.valid = True
            self.observer.clear()
        except Exception as e:
            # print("Error in UpdatePipeline :", e)
            # traceback.print_stack()
            print(e)
            self.valid = False

        return self.valid

    def LoadVariables(self, vars):
        if not self.valid:
            return
        self.ctrl_data.Variables = vars
        self.test_data.Variables = vars


if __name__ == "__main__":
    e = EAMVisSource()



from trame.decorators import change
from trame.widgets import html
from trame.widgets import vuetify3 as v3

from e3sm_compareview import __version__ as app_version
from e3sm_compareview.assets import ASSETS
from e3sm_quickview.components import css, tools
from e3sm_quickview.utils import constants, js


class AppLogo(v3.VTooltip):
    def __init__(self, compact="compact_drawer"):
        super().__init__(
            text=f"CompareView {app_version}",
            disabled=(f"!{compact}",),
        )
        with self:
            with v3.Template(v_slot_activator="{ props }"):
                with v3.VListItem(
                    v_bind="props",
                    click=f"{compact} = !{compact}",
                ):
                    with html.Div(classes="d-flex align-center flex-grow-1"):
                        html.Img(
                            src=ASSETS.icon,
                            v_if=compact,
                            style="width: 24px; height: 24px; object-fit: contain;",
                        )
                        html.Img(
                            src=ASSETS.logo,
                            v_else=True,
                            style="height: 28px; max-width: 160px; object-fit: contain;",
                        )
                    v3.VProgressCircular(
                        color="primary",
                        indeterminate=True,
                        v_show="trame__busy",
                        v_if=compact,
                        style="position: absolute !important;left: 50%;top: 50%; transform: translate(-50%, -50%);",
                    )
                    v3.VProgressLinear(
                        v_else=True,
                        color="primary",
                        indeterminate=True,
                        v_show="trame__busy",
                        absolute=True,
                        style="top:90%;width:100%;",
                    )


class Tools(v3.VNavigationDrawer):
    def __init__(self, reset_camera=None):
        super().__init__(
            permanent=True,
            rail=("compact_drawer", True),
            width=253,
            style="transform: none;",
        )

        with self:
            with html.Div(style=css.NAV_BAR_TOP):
                with v3.VList(
                    density="compact",
                    nav=True,
                    select_strategy="independent",
                    v_model_selected=("active_tools", ["load-data"]),
                ):
                    AppLogo()
                    tools.ResetCamera(click=reset_camera)

                    v3.VDivider(classes="my-1")  # ---------------------

                    tools.StateImportExport()
                    tools.OpenFile()

                    v3.VDivider(classes="my-1")  # ---------------------

                    tools.FieldSelection()
                    tools.DataSelection()
                    tools.Animation()

                    v3.VDivider(classes="my-1")  # ---------------------

                    tools.LayoutManagement()
                    tools.MapProjection()
                    tools.Cropping()

                    # dev add-on ui reload
                    if self.server.hot_reload:
                        v3.VDivider(classes="my-1")  # ---------------------
                        tools.ActionButton(
                            compact="compact_drawer",
                            title="Refresh UI",
                            icon="mdi-database-refresh-outline",
                            click=self.ctrl.on_server_reload,
                        )

            with html.Div(style=css.NAV_BAR_BOTTOM):
                v3.VDivider()
                v3.VLabel(
                    f"{app_version}",
                    classes="text-center text-caption d-block text-wrap",
                )


class FieldSelection(v3.VNavigationDrawer):
    def __init__(self, load_variables=None):
        super().__init__(
            model_value=(js.is_active("select-fields"),),
            width=500,
            permanent=True,
            style=(f"{js.is_active('select-fields')} ? 'transform: none;' : ''",),
        )

        with self:
            with html.Div(style="position:fixed;top:0;width: 500px;"):
                # Simulation file selection dropdowns
                with v3.VRow(classes="mx-2 my-2", dense=True, v_if="pv_files_data_simulation_files.length > 1"):
                    with v3.VCol(cols=6):
                        v3.VSelect(
                            v_model=("ctrl_simulation_file", "pv_files_data_simulation_files[0]"),
                            items=("pv_files_data_simulation_files", []),
                            label="Control Simulation",
                            density="compact",
                            variant="outlined",
                            hide_details=True,
                        )
                    with v3.VCol(cols=6):
                        v3.VSelect(
                            v_model=("test_simulation_file", "pv_files_data_simulation_files[1] || pv_files_data_simulation_files[0]"),
                            items=("pv_files_data_simulation_files", []),
                            label="Test Simulation",
                            density="compact",
                            variant="outlined",
                            hide_details=True,
                        )

                with v3.VCardActions(classes="px-2", style="min-height: 0;"):
                    v3.VBtn(
                        classes="text-none",
                        color="primary",
                        prepend_icon="mdi-database",
                        text=(
                            "`Load ${variables_selected.length} variable${variables_selected.length > 1 ? 's' :''}`",
                        ),
                        variant="flat",
                        block=True,
                        disabled=(
                            "variables_selected.length === 0 || variables_loaded",
                        ),
                        click=load_variables,
                    )
                with v3.VCardActions(
                    classes="flex-wrap py-1",
                    style="overflow-y: auto; max-height: 100px; min-height: 42px;",
                ):
                    v3.VChip(
                        "{{ variables_selected.filter(id => variables_listing.find(v => v.id === id)?.type === vtype.name).length }} {{ vtype.name }}",
                        v_for="(vtype, idx) in variable_types",
                        key="idx",
                        color=("vtype.color",),
                        v_show=(
                            "variables_selected.filter(id => variables_listing.find(v => v.id === id)?.type === vtype.name).length",
                        ),
                        size="small",
                        closable=True,
                        click_close=(
                            "variables_selected = variables_selected.filter(id => variables_listing.find(v => v.id === id)?.type !== vtype.name)"
                        ),
                        classes="ma-1",
                    )

                v3.VTextField(
                    v_model=("variables_filter", ""),
                    hide_details=True,
                    color="primary",
                    placeholder="Filter",
                    density="compact",
                    variant="outlined",
                    classes="mx-2",
                    prepend_inner_icon="mdi-magnify",
                    clearable=True,
                )
                with html.Div(style="margin:1px;"):
                    v3.VDataTable(
                        v_model=("variables_selected", []),
                        show_select=True,
                        item_value="id",
                        density="compact",
                        fixed_header=True,
                        headers=(
                            "variables_headers",
                            constants.VAR_HEADERS,
                        ),
                        items=("variables_listing", []),
                        height="calc(100vh - 6rem)",
                        style="user-select: none; cursor: pointer;",
                        hover=True,
                        search=("variables_filter", ""),
                        items_per_page=-1,
                        hide_default_footer=True,
                    )

    @change("variables_selected")
    def _on_dirty_variable_selection(self, **_):
        self.state.variables_loaded = False

    @change("ctrl_simulation_file", "test_simulation_file")
    def _on_dirty_simulation_selection(self, **_):
        self.state.variables_loaded = False



import math

from trame.app import TrameComponent
from trame.ui.html import DivLayout
from trame.widgets import paraview as pvw, vuetify3 as v3, client, html
from trame.decorators import controller

from trame_dataclass.core import StateDataModel

from paraview import simple

from e3sm_compareview.components import view as tview
from e3sm_quickview.presets import COLOR_BLIND_SAFE
from e3sm_quickview.utils.color import COLORBAR_CACHE, get_cached_colorbar_image


def auto_size_to_col(size):
    if size == 1:
        return 12

    if size >= 8 and size % 2 == 0:
        return 3

    if size % 3 == 0:
        return 4

    if size % 2 == 0:
        return 6

    return auto_size_to_col(size + 1)


COL_SIZE_LOOKUP = {
    0: auto_size_to_col,
    1: 12,
    2: 6,
    3: 4,
    4: 3,
    6: 2,
    12: 1,
    "flow": None,
}


def lut_name(element):
    return element.get("name").lower()


class ViewConfiguration(StateDataModel):
    variable: str
    preset: str = "Inferno (matplotlib)"
    preset_img: str
    invert: bool = False
    color_blind: bool = False
    use_log_scale: bool = False
    color_value_min: str = "0"
    color_value_max: str = "1"
    color_value_min_valid: bool = True
    color_value_max_valid: bool = True
    color_range: list[float] = (0, 1)
    override_range: bool = False
    order: int = 0
    size: int = 4
    offset: int = 0
    break_row: bool = False
    menu: bool = False
    swap_group: list[str]
    search: str | None


class VariableView(TrameComponent):
    def __init__(self, server, source, variable_name, variable_type):
        super().__init__(server)
        self.config = ViewConfiguration(server, variable=variable_name)
        self.source = source
        self.variable_name = variable_name
        self.variable_type = variable_type
        self.name = f"view_{self.variable_name}"
        if self.variable_name.endswith(("_ctrl", "_test")):
            self.config.preset = "navia"
        elif self.variable_name.endswith("_diff"):
            self.config.preset = "Cool to Warm (Extended)"
        elif self.variable_name.endswith(("_comp1", "_comp2")):
            self.config.preset = "bam"
            self.config.invert = True
        self.view = simple.CreateRenderView()
        self.view.GetRenderWindow().SetOffScreenRendering(True)
        self.view.InteractionMode = "2D"
        self.view.OrientationAxesVisibility = 0
        self.view.UseColorPaletteForBackground = 0
        self.view.BackgroundColorMode = "Single Color"
        self.view.Background = [1, 1, 1]
        self.view.Background2 = [1, 1, 1]
        self.view.CameraParallelProjection = 1
        self.view.Size = 0  # make the interactive widget non responsive
        self.representation = simple.Show(
            proxy=source.views["atmosphere_data"],
            view=self.view,
        )

        # Lookup table color management
        simple.ColorBy(self.representation, ("CELLS", variable_name))
        self.lut = simple.GetColorTransferFunction(variable_name)
        self.lut.NanOpacity = 0.0

        self.view.ResetActiveCameraToNegativeZ()
        self.view.ResetCamera(True, 0.9)
        self.disable_render = False

        # Add annotation to the view
        # - continents
        globe = source.views["continents"]
        repG = simple.Show(globe, self.view)
        simple.ColorBy(repG, None)
        repG.SetRepresentationType("Wireframe")
        repG.RenderLinesAsTubes = 1
        repG.LineWidth = 1.0
        repG.AmbientColor = [0.67, 0.67, 0.67]
        repG.DiffuseColor = [0.67, 0.67, 0.67]
        self.rep_globe = repG
        # - gridlines
        annot = source.views["grid_lines"]
        repAn = simple.Show(annot, self.view)
        repAn.SetRepresentationType("Wireframe")
        repAn.AmbientColor = [0.67, 0.67, 0.67]
        repAn.DiffuseColor = [0.67, 0.67, 0.67]
        repAn.Opacity = 0.4
        self.rep_grid = repAn

        # Reactive behavior
        self.config.watch(
            ["color_value_min", "color_value_max"],
            self.color_range_str_to_float,
        )
        self.config.watch(
            ["override_range", "color_range"], self.update_color_range, eager=True
        )
        self.config.watch(
            ["preset", "invert", "use_log_scale"], self.update_color_preset, eager=True
        )

        # GUI
        self._build_ui()

    def render(self):
        if self.disable_render or not self.ctx.has(self.name):
            return
        self.ctx[self.name].update()

    def set_camera_modified(self, fn):
        self._observer = self.camera.AddObserver("ModifiedEvent", fn)

    @property
    def camera(self):
        return self.view.GetActiveCamera()

    def reset_camera(self):
        self.view.InteractionMode = "2D"
        self.view.ResetActiveCameraToNegativeZ()
        self.view.ResetCamera(True, 0.9)
        self.ctx[self.name].update()

    def update_color_preset(self, name, invert, log_scale):
        self.config.preset = name
        self.config.preset_img = get_cached_colorbar_image(
            self.config.preset,
            self.config.invert,
        )
        self.lut.ApplyPreset(self.config.preset, True)
        if invert:
            self.lut.InvertTransferFunction()
        if log_scale:
            self.lut.MapControlPointsToLogSpace()
            self.lut.UseLogScale = 1
        self.render()

    def color_range_str_to_float(self, color_value_min, color_value_max):
        try:
            min_value = float(color_value_min)
            self.config.color_value_min_valid = not math.isnan(min_value)
        except ValueError:
            self.config.color_value_min_valid = False

        try:
            max_value = float(color_value_max)
            self.config.color_value_max_valid = not math.isnan(max_value)
        except ValueError:
            self.config.color_value_max_valid = False

        if self.config.color_value_min_valid and self.config.color_value_max_valid:
            self.config.color_range = [min_value, max_value]

    def _parse_comparison(self):
        for comp_type in ("ctrl", "test", "diff", "comp1", "comp2"):
            suffix = f"_{comp_type}"
            if self.variable_name.endswith(suffix):
                return self.variable_name[: -len(suffix)], comp_type
        return self.variable_name, None

    @staticmethod
    def _is_finite_range(data_range):
        if data_range is None or len(data_range) < 2:
            return False
        return math.isfinite(data_range[0]) and math.isfinite(data_range[1])

    def _ctrl_test_visible(self):
        selected_columns = getattr(self.state, "selected_columns", None)
        if not selected_columns:
            return False
        return "ctrl" in selected_columns and "test" in selected_columns

    def _get_ctrl_test_combined_range(self):
        base_name, comp_type = self._parse_comparison()
        if comp_type not in ("ctrl", "test"):
            return None
        if not self._ctrl_test_visible():
            return None

        data_info = self.source.views["atmosphere_data"].GetCellDataInformation()
        ctrl_array = data_info.GetArray(f"{base_name}_ctrl")
        test_array = data_info.GetArray(f"{base_name}_test")
        if not ctrl_array or not test_array:
            return None

        ctrl_range = ctrl_array.GetRange()
        test_range = test_array.GetRange()
        if not (self._is_finite_range(ctrl_range) and self._is_finite_range(test_range)):
            return None

        return [min(ctrl_range[0], test_range[0]), max(ctrl_range[1], test_range[1])]

    def _get_diff_centered_range(self):
        _, comp_type = self._parse_comparison()
        if comp_type != "diff":
            return None

        data_info = self.source.views["atmosphere_data"].GetCellDataInformation()
        diff_array = data_info.GetArray(self.variable_name)
        if not diff_array:
            return None

        diff_range = diff_array.GetRange()
        if not self._is_finite_range(diff_range):
            return None

        max_abs = max(abs(diff_range[0]), abs(diff_range[1]))
        return [-max_abs, max_abs]

    def _comp_visible(self):
        selected_columns = getattr(self.state, "selected_columns", None)
        if not selected_columns:
            return False
        return "comp1" in selected_columns and "comp2" in selected_columns

    def _get_comp_centered_range(self):
        base_name, comp_type = self._parse_comparison()
        if comp_type not in ("comp1", "comp2"):
            return None

        data_info = self.source.views["atmosphere_data"].GetCellDataInformation()
        if self._comp_visible():
            comp1_array = data_info.GetArray(f"{base_name}_comp1")
            comp2_array = data_info.GetArray(f"{base_name}_comp2")
            if not comp1_array or not comp2_array:
                return None

            comp1_range = comp1_array.GetRange()
            comp2_range = comp2_array.GetRange()
            if not (
                self._is_finite_range(comp1_range)
                and self._is_finite_range(comp2_range)
            ):
                return None

            max_abs = max(
                abs(comp1_range[0]),
                abs(comp1_range[1]),
                abs(comp2_range[0]),
                abs(comp2_range[1]),
            )
            return [-max_abs, max_abs]

        comp_array = data_info.GetArray(self.variable_name)
        if not comp_array:
            return None

        comp_range = comp_array.GetRange()
        if not self._is_finite_range(comp_range):
            return None

        max_abs = max(abs(comp_range[0]), abs(comp_range[1]))
        return [-max_abs, max_abs]

    def update_color_range(self, *_):
        if self.config.override_range:
            skip_update = False
            if math.isnan(self.config.color_range[0]):
                skip_update = True
                self.config.color_value_min_valid = False

            if math.isnan(self.config.color_range[1]):
                skip_update = True
                self.config.color_value_max_valid = False

            if skip_update:
                return

            self.lut.RescaleTransferFunction(*self.config.color_range)
        else:
            data_range = self._get_diff_centered_range()
            if data_range is None:
                data_range = self._get_comp_centered_range()
            if data_range is None:
                data_range = self._get_ctrl_test_combined_range()
            if data_range is None:
                self.representation.RescaleTransferFunctionToDataRange(False, True)
                data_array = (
                    self.source.views["atmosphere_data"]
                    .GetCellDataInformation()
                    .GetArray(self.variable_name)
                )
                if data_array:
                    data_range = data_array.GetRange()

            if data_range is not None:
                self.config.color_range = data_range
                self.config.color_value_min = str(data_range[0])
                self.config.color_value_max = str(data_range[1])
                self.config.color_value_min_valid = True
                self.config.color_value_max_valid = True
                self.lut.RescaleTransferFunction(*data_range)
        self.render()

    def _build_ui(self):
        with DivLayout(
            self.server, template_name=self.name, connect_parent=False, classes="h-100"
        ) as self.ui:
            self.ui.root.classes = "h-100"
            with v3.VCard(
                variant="tonal",
                style=(
                    "active_layout !== 'auto_layout' ? `height: calc(100% - ${top_padding}px;` : 'overflow-hidden'",
                ),
                tile=("active_layout !== 'auto_layout'",),
            ):
                with v3.VRow(
                    dense=True,
                    classes="ma-0 pa-0 bg-white text-black d-flex align-center border-b-thin",
                    style="flex-wrap: nowrap;",
                ):
                    tview.create_size_menu(self.name, self.config)
                    html.Div(
                        self.variable_name,
                        classes="text-subtitle-2 pr-2",
                        style="user-select: none; overflow: hidden; text-overflow: ellipsis; white-space: nowrap; min-width: 0;",
                    )

                    v3.VIcon(
                        "mdi-lock-outline",
                        size="x-small",
                        v_show=("lock_views", False),
                        style="transform: scale(0.75);",
                    )

                    v3.VSpacer()
                    html.Div(
                        "t = {{ time_idx }}",
                        classes="text-caption px-1",
                        v_if="timestamps.length > 1",
                    )
                    if self.variable_type == "m":
                        html.Div(
                            "[k = {{ midpoint_idx }}]",
                            classes="text-caption px-1",
                            v_if="midpoints.length > 1",
                        )
                    if self.variable_type == "i":
                        html.Div(
                            "[k = {{ interface_idx }}]",
                            classes="text-caption px-1",
                            v_if="interfaces.length > 1",
                        )
                    # v3.VSpacer()
                    # html.Div(
                    #     "avg = {{"
                    #     f"fields_avgs['{self.variable_name}']?.toExponential(2) || 'N/A'"
                    #     "}}",
                    #     classes="text-caption px-1",
                    # )

                with html.Div(
                    style=(
                        """
                        {
                            aspectRatio: active_layout === 'auto_layout' ? aspect_ratio : null,
                            height: active_layout !== 'auto_layout' ? 'calc(100% - 2.4rem)' : null,
                            pointerEvents: lock_views ? 'none': null,
                        }
                        """,
                    ),
                ):
                    pvw.VtkRemoteView(
                        self.view, interactive_ratio=1, ctx_name=self.name
                    )

                tview.create_bottom_bar(self.config, self.update_color_preset)


class ViewManager(TrameComponent):
    def __init__(self, server, source):
        super().__init__(server)
        self.source = source
        self._var2view = {}
        self._camera_sync_in_progress = False
        self._last_vars = {}
        self._active_configs = {}

        pvw.initialize(self.server)

        self.state.luts_normal = [
            {"name": k, "url": v["normal"], "safe": k in COLOR_BLIND_SAFE}
            for k, v in COLORBAR_CACHE.items()
        ]
        self.state.luts_inverted = [
            {"name": k, "url": v["inverted"], "safe": k in COLOR_BLIND_SAFE}
            for k, v in COLORBAR_CACHE.items()
        ]

        # Sort lists
        self.state.luts_normal.sort(key=lut_name)
        self.state.luts_inverted.sort(key=lut_name)

    def refresh_ui(self, **_):
        for view in self._var2view.values():
            view._build_ui()

    def reset_camera(self):
        views = list(self._var2view.values())
        for view in views:
            view.disable_render = True

        for view in views:
            view.reset_camera()

        for view in views:
            view.disable_render = False

    def render(self):
        for view in list(self._var2view.values()):
            view.render()

    def update_color_range(self):
        for view in list(self._var2view.values()):
            view.update_color_range()

    def get_view(self, variable_name, variable_type):
        view = self._var2view.get(variable_name)
        if view is None:
            view = self._var2view.setdefault(
                variable_name,
                VariableView(self.server, self.source, variable_name, variable_type),
            )
            view.set_camera_modified(self.sync_camera)

        return view

    def sync_camera(self, camera, *_):
        if self._camera_sync_in_progress:
            return
        self._camera_sync_in_progress = True

        for var_view in self._var2view.values():
            cam = var_view.camera
            if cam is camera:
                continue
            cam.DeepCopy(camera)
            var_view.render()

        self._camera_sync_in_progress = False

    @controller.set("swap_variables")
    def swap_variable(self, variable_a, variable_b):
        config_a = self._active_configs[variable_a]
        config_b = self._active_configs[variable_b]
        config_a.order, config_b.order = config_b.order, config_a.order
        config_a.size, config_b.size = config_b.size, config_a.size
        config_a.offset, config_b.offset = config_b.offset, config_a.offset
        config_a.break_row, config_b.break_row = config_b.break_row, config_a.break_row

    def apply_size(self, n_cols):
        if not self._last_vars:
            return

        if n_cols == 0:
            # Auto based on group size
            if self.state.layout_grouped:
                for var_type in "smi":
                    var_names = self._last_vars[var_type]
                    total_size = len(var_names)

                    if total_size == 0:
                        continue

                    size = auto_size_to_col(total_size)
                    for name in var_names:
                        config = self.get_view(name, var_type).config
                        config.size = size

            else:
                size = auto_size_to_col(len(self._active_configs))
                for config in self._active_configs.values():
                    config.size = size
        else:
            # uniform size
            for config in self._active_configs.values():
                config.size = COL_SIZE_LOOKUP[n_cols]

    def build_auto_layout(self, variables=None):
        if variables is None:
            variables = self._last_vars

        self._last_vars = variables

        # Create UI based on variables
        self.state.swap_groups = {}
        # Build a lookup from type name to color from state.variable_types
        type_to_color = {vt["name"]: vt["color"] for vt in self.state.variable_types}
        with DivLayout(self.server, template_name="auto_layout") as self.ui:
            if self.state.layout_grouped:
                with v3.VCol(classes="pa-1"):
                    for var_type in variables.keys():
                        var_names = variables[var_type]
                        total_size = len(var_names)
                        print(total_size, var_names, var_type)

                        if total_size == 0:
                            continue

                        for var_name in var_names:
                            # Define all possible comparison types
                            all_comp_types = ["ctrl", "test", "diff", "comp1", "comp2"]
                            # Build the full list of comparisons
                            var_comps = [f"{var_name}_{comp_type}" for comp_type in all_comp_types]

                            # Look up color from variable_types to match chip colors
                            border_color = type_to_color.get(str(var_type), "primary")
                            with v3.VAlert(
                                border="start",
                                classes="pr-1 py-1 pl-3 mb-1",
                                variant="flat",
                                border_color=border_color,
                            ):
                                with v3.VRow(dense=True):
                                    for comp_type, name in zip(all_comp_types, var_comps):
                                        print("Creating view for", name)
                                        view = self.get_view(name, var_type)
                                        view.config.swap_group = sorted(
                                            [n for n in var_comps if n != name]
                                        )
                                        with view.config.provide_as("config"):
                                            # Only show column if it's in selected_columns
                                            with v3.Template(v_if=f"selected_columns.includes('{comp_type}')"):
                                                v3.VCol(
                                                    v_if="config.break_row",
                                                    cols=12,
                                                    classes="pa-0",
                                                    style=("`order: ${config.order};`",),
                                                )
                                                # For flow handling
                                                with v3.Template(v_if="!config.size"):
                                                    v3.VCol(
                                                        v_for="i in config.offset",
                                                        key="i",
                                                        style=("{ order: config.order }",),
                                                    )
                                                with v3.VCol(
                                                    offset=("config.offset * config.size",),
                                                    cols=("Math.floor(12 / selected_columns.length)",),
                                                    # cols=("config.size",),
                                                    style=("`order: ${config.order};`",),
                                                ):
                                                    client.ServerTemplate(name=view.name)
            else:
                all_names = [name for names in variables.values() for name in names]
                with v3.VRow(dense=True, classes="pa-2"):
                    for var_type in variables.keys():
                        var_names = variables[var_type]
                        for name in var_names:
                            view = self.get_view(name, var_type)
                            view.config.swap_group = sorted(
                                [n for n in all_names if n != name]
                            )
                            with view.config.provide_as("config"):
                                v3.VCol(
                                    v_if="config.break_row",
                                    cols=12,
                                    classes="pa-0",
                                    style=("`order: ${config.order};`",),
                                )

                                # For flow handling
                                with v3.Template(v_if="!config.size"):
                                    v3.VCol(
                                        v_for="i in config.offset",
                                        key="i",
                                        style=("{ order: config.order }",),
                                    )
                                with v3.VCol(
                                    offset=(
                                        "config.size ? config.offset * config.size : 0",
                                    ),
                                    cols=("config.size",),
                                    style=("`order: ${config.order};`",),
                                ):
                                    client.ServerTemplate(name=view.name)

        # Assign any missing order
        self._active_configs = {}
        existed_order = set()
        order_max = 0
        orders_to_update = []
        for var_type in variables.keys():
            var_names = variables[var_type]
            for var_name in var_names:
                # Define all possible comparison types
                all_comp_types = ["ctrl", "test", "diff", "comp1", "comp2"]
                # Build the full list of comparisons
                var_comps = [f"{var_name}_{comp_type}" for comp_type in all_comp_types]

                for name in var_comps:
                    config = self.get_view(name, var_type).config
                    self._active_configs[name] = config
                    if config.order:
                        order_max = max(order_max, config.order)
                        assert config.order not in existed_order, "Order already assigned"
                        existed_order.add(config.order)
                    else:
                        orders_to_update.append(config)

        next_order = order_max + 1
        for config in orders_to_update:
            config.order = next_order
            next_order += 1



from trame.widgets import html, vuetify3 as v3


def create_size_menu(name, config):
    with v3.VBtn(
        icon=True,
        density="compact",
        variant="plain",
        classes="mx-1",
        size="small",
    ):
        v3.VIcon(
            "mdi-arrow-expand",
            size="x-small",
            style="transform: scale(-1, 1);",
        )
        with v3.VMenu(activator="parent"):
            with config.provide_as("config"):
                with v3.VList(density="compact"):
                    v3.VListItem(
                        subtitle="Full Screen",
                        click=f"active_layout = '{name}'",
                    )
                    v3.VDivider()
                    with v3.VListItem(
                        subtitle="Line Break",
                        click="config.break_row = !config.break_row",
                    ):
                        with v3.Template(v_slot_append=True):
                            v3.VSwitch(
                                v_model="config.break_row",
                                hide_details=True,
                                density="compact",
                                color="primary",
                            )
                    with v3.VListItem(subtitle="Offset"):
                        v3.VBtn(
                            "0",
                            classes="text-none ml-2",
                            size="small",
                            variant="outined",
                            click="config.offset = 0",
                            active=("config.offset === 0",),
                        )
                        v3.VBtn(
                            "1",
                            classes="text-none ml-2",
                            size="small",
                            variant="outined",
                            click="config.offset = 1",
                            active=("config.offset === 1",),
                        )
                        v3.VBtn(
                            "2",
                            classes="text-none ml-2",
                            size="small",
                            variant="outined",
                            click="config.offset = 2",
                            active=("config.offset === 2",),
                        )
                        v3.VBtn(
                            "3",
                            classes="text-none ml-2",
                            size="small",
                            variant="outined",
                            click="config.offset = 3",
                            active=("config.offset === 3",),
                        )
                        v3.VBtn(
                            "4",
                            classes="text-none ml-2",
                            size="small",
                            variant="outined",
                            click="config.offset = 4",
                            active=("config.offset === 4",),
                        )
                        v3.VBtn(
                            "5",
                            classes="text-none ml-2",
                            size="small",
                            variant="outined",
                            click="config.offset = 5",
                            active=("config.offset === 5",),
                        )
                    v3.VDivider()

                    v3.VListItem(
                        subtitle="Full width",
                        click="active_layout = 'auto_layout';config.size = 12",
                    )
                    v3.VListItem(
                        subtitle="1/2 width",
                        click="active_layout = 'auto_layout';config.size = 6",
                    )
                    v3.VListItem(
                        subtitle="1/3 width",
                        click="active_layout = 'auto_layout';config.size = 4",
                    )
                    v3.VListItem(
                        subtitle="1/4 width",
                        click="active_layout = 'auto_layout';config.size = 3",
                    )
                    v3.VListItem(
                        subtitle="1/6 width",
                        click="active_layout = 'auto_layout';config.size = 2",
                    )


def create_bottom_bar(config, update_color_preset):
    with config.provide_as("config"):
        with html.Div(
            classes="bg-white text-black d-flex align-center border-t-thin",
            style="height:1rem;position:relative;top:0;user-select:none;cursor:context-menu;",
        ):
            with v3.VMenu(
                v_model="config.menu",
                activator="parent",
                location=(
                    "active_layout !== 'auto_layout' || config.size == 12 ? 'top' : 'end'",
                ),
                close_on_content_click=False,
            ):
                with v3.VCard(style="max-width: 360px;min-width: 360px;"):
                    with v3.VCardItem(classes="py-0 px-2"):
                        with html.Div(classes="d-flex align-center"):
                            v3.VIconBtn(
                                raw_attrs=[
                                    '''v-tooltip:bottom="config.color_blind ? 'Colorblind safe presets' : 'All color presets'"'''
                                ],
                                icon=(
                                    "config.color_blind ? 'mdi-shield-check-outline' : 'mdi-palette'",
                                ),
                                click="config.color_blind = !config.color_blind",
                                size="small",
                                text="Colorblind safe",
                                variant="text",
                            )
                            v3.VIconBtn(
                                raw_attrs=[
                                    '''v-tooltip:bottom="config.invert ? 'Inverted preset' : 'Normal preset'"'''
                                ],
                                icon=(
                                    "config.invert ? 'mdi-invert-colors' : 'mdi-invert-colors-off'",
                                ),
                                click="config.invert = !config.invert",
                                size="small",
                                text="Invert",
                                variant="text",
                            )
                            v3.VIconBtn(
                                raw_attrs=[
                                    '''v-tooltip:bottom="config.use_log_scale ? 'Use log scale' : 'Use linear scale'"'''
                                ],
                                icon=(
                                    "config.use_log_scale ? 'mdi-math-log' : 'mdi-stairs'",
                                ),
                                click="config.use_log_scale = !config.use_log_scale",
                                size="small",
                                text=(
                                    "config.use_log_scale ? 'Log scale' : 'Linear scale'",
                                ),
                                variant="text",
                            )
                            v3.VIconBtn(
                                raw_attrs=[
                                    '''v-tooltip:bottom="config.override_range ? 'Use custom range' : 'Use data range'"'''
                                ],
                                icon=(
                                    "config.override_range ? 'mdi-arrow-expand-horizontal' : 'mdi-pencil'",
                                ),
                                click="config.override_range = !config.override_range",
                                size="small",
                                text="Use data range",
                                variant="text",
                            )
                            v3.VTextField(
                                v_model="config.search",
                                clearable=True,
                                placeholder=("config.preset",),
                                click_clear="config.search = null",
                                single_line=True,
                                variant="solo",
                                density="compact",
                                flat=True,
                                hide_details="auto",
                                # style="min-width: 150px;",
                                classes="d-inline",
                                reverse=True,
                            )
                            v3.VIconBtn(
                                icon="mdi-close",
                                size="small",
                                text="Close",
                                click="config.menu=false",
                            )

                    with v3.VCardItem(v_show="config.override_range", classes="py-0"):
                        v3.VTextField(
                            v_model="config.color_value_min",
                            hide_details=True,
                            density="compact",
                            variant="outlined",
                            flat=True,
                            label="Min",
                            classes="mt-2",
                            error=("!config.color_value_min_valid",),
                        )
                        v3.VTextField(
                            v_model="config.color_value_max",
                            hide_details=True,
                            density="compact",
                            variant="outlined",
                            flat=True,
                            label="Max",
                            classes="mt-2",
                            error=("!config.color_value_max_valid",),
                        )
                    v3.VDivider()
                    with v3.VList(density="compact", max_height="40vh"):
                        with v3.VListItem(
                            v_for="entry in (config.invert ? luts_inverted : luts_normal)",
                            v_show="(config.search?.length ? entry.name.toLowerCase().includes(config.search.toLowerCase()) : 1) && (!config.color_blind || entry.safe)",
                            key="entry.name",
                            subtitle=("entry.name",),
                            click=(
                                update_color_preset,
                                "[entry.name, config.invert, config.use_log_scale]",
                            ),
                            active=("config.preset === entry.name",),
                        ):
                            html.Img(
                                src=("entry.url",),
                                style="width:100%;min-width:20rem;height:1rem;",
                                classes="rounded",
                            )
            html.Div(
                "{{ utils.quickview.formatRange(config.color_range?.[0], config.use_log_scale) }}",
                classes="text-caption px-2 text-no-wrap",
            )
            with html.Div(classes="overflow-hidden rounded w-100", style="height:70%;"):
                html.Img(
                    src=("config.preset_img",),
                    style="width:100%;height:2rem;",
                    draggable=False,
                )
            html.Div(
                "{{ utils.quickview.formatRange(config.color_range?.[1], config.use_log_scale) }}",
                classes="text-caption px-2 text-no-wrap",
            )



import asyncio
import datetime
import json
import os
from pathlib import Path

from trame.app import TrameApp, asynchronous, file_upload
from trame.decorators import change, controller, life_cycle, trigger
from trame.ui.vuetify3 import VAppLayout
from trame.widgets import client, dataclass, html, tauri
from trame.widgets import trame as tw
from trame.widgets import vuetify3 as v3

from e3sm_compareview.assets import ASSETS
from e3sm_compareview.components import drawers, file_browser, toolbars
from e3sm_compareview.pipeline import EAMVisSource
from e3sm_compareview.view_manager import ViewManager
from e3sm_quickview.components import css, dialogs, doc
from e3sm_quickview import module as qv_module
from e3sm_quickview.utils import cli, compute

v3.enable_lab()


class EAMApp(TrameApp):
    def __init__(self, server=None):
        super().__init__(server)

        # Pre-load deferred widgets
        dataclass.initialize(self.server)
        self.server.enable_module(qv_module)

        # CLI
        args = cli.configure_and_parse(self.server.cli)

        # Initial UI state
        self.state.update(
            {
                "trame__title": "CompareView",
                "trame__favicon": ASSETS.icon,
                "is_tauri": False,
                "animation_play": False,
                # All available variables
                "variables_listing": [],
                # Selected variables to load
                "variables_selected": [],
                # Control 'Load Variables' button availability
                "variables_loaded": False,
                # Dynamic type-color mapping (populated when data loads)
                "variable_types": [],
                # Dimension arrays (will be populated dynamically)
                "midpoints": [],
                "interfaces": [],
                "timestamps": [],
                # Fields summaries
                "fields_avgs": {},
                # Simulation file selection for ctrl/test comparison
                "ctrl_simulation_file": "",
                "test_simulation_file": "",
                # Column visibility for variable comparisons
                "selected_columns": ["ctrl", "test", "diff", "comp1", "comp2"],
            }
        )

        # Data input
        self.source = EAMVisSource()

        # Helpers
        self.view_manager = ViewManager(self.server, self.source)
        self.file_browser = file_browser.ParaViewFileBrowser(
            self.server,
            prefix="pv_files",
            home=None if args.user_home else args.workdir,  # can use current=
            group="",
        )

        # Process CLI to pre-load data
        if args.state is not None:
            state_content = json.loads(Path(args.state).read_text())

            async def wait_for_import(**_):
                await self.import_state(state_content)

            self.ctrl.on_server_ready.add_task(wait_for_import)
        elif args.data and args.conn:
            self.file_browser.set_data_simulation(args.data)
            self.file_browser.set_data_connectivity(args.conn)
            self.ctrl.on_server_ready.add(self.file_browser.load_data_files)

        # Development setup
        if self.server.hot_reload:
            self.ctrl.on_server_reload.add(self._build_ui)
            self.ctrl.on_server_reload.add(self.view_manager.refresh_ui)

        # GUI
        self._build_ui()

    # -------------------------------------------------------------------------
    # Tauri adapter
    # -------------------------------------------------------------------------

    @life_cycle.server_ready
    def _tauri_ready(self, **_):
        os.write(1, f"tauri-server-port={self.server.port}\n".encode())

    @life_cycle.client_connected
    def _tauri_show(self, **_):
        os.write(1, "tauri-client-ready\n".encode())

    # -------------------------------------------------------------------------
    # UI definition
    # -------------------------------------------------------------------------

    def _build_ui(self, **_):
        with VAppLayout(self.server, fill_height=True) as self.ui:
            # Keyboard shortcut
            with tw.MouseTrap(
                ResetCamera=self.view_manager.reset_camera,
                SizeAuto=(self.view_manager.apply_size, "[0]"),
                Size1=(self.view_manager.apply_size, "[1]"),
                Size2=(self.view_manager.apply_size, "[2]"),
                Size3=(self.view_manager.apply_size, "[3]"),
                Size4=(self.view_manager.apply_size, "[4]"),
                Size6=(self.view_manager.apply_size, "[6]"),
                SizeFlow=(self.view_manager.apply_size, "['flow']"),
                ToolbarLayout=(self.toggle_toolbar, "['adjust-layout']"),
                ToolbarCrop=(self.toggle_toolbar, "['adjust-databounds']"),
                ToolbarSelect=(self.toggle_toolbar, "['select-slice-time']"),
                ToolbarAnimation=(self.toggle_toolbar, "['animation-controls']"),
                ToggleVariableSelection=(self.toggle_toolbar, "['select-fields']"),
                RemoveAllToolbars=(self.toggle_toolbar),
                ToggleGroups="layout_grouped = !layout_grouped",
                ProjectionEquidistant="projection = ['Cyl. Equidistant']",
                ProjectionRobinson="projection = ['Robinson']",
                ProjectionMollweide="projection = ['Mollweide']",
                ToggleViewLock="lock_views = !lock_views",
                FileOpen=(self.toggle_toolbar, "['load-data']"),
                SaveState="trigger('download_state')",
                UploadState="utils.get('document').querySelector('#fileUpload').click()",
                ToggleHelp="compact_drawer = !compact_drawer",
            ) as mt:
                mt.bind(["z"], "ResetCamera")
                mt.bind(["alt+0", "0"], "SizeAuto")
                mt.bind(["alt+1", "1"], "Size1")
                mt.bind(["alt+2", "2"], "Size2")
                mt.bind(["alt+3", "3"], "Size3")
                mt.bind(["alt+4", "4"], "Size4")
                mt.bind(["alt+6", "6"], "Size6")
                mt.bind(["="], "SizeFlow")

                mt.bind("c", "ProjectionEquidistant")
                mt.bind("r", "ProjectionRobinson")
                mt.bind("m", "ProjectionMollweide")

                mt.bind("f", "FileOpen")
                mt.bind("d", "SaveState")
                mt.bind("u", "UploadState")
                mt.bind("h", "ToggleHelp")

                mt.bind("p", "ToolbarLayout")
                mt.bind("l", "ToolbarCrop")
                mt.bind("s", "ToolbarSelect")
                mt.bind("a", "ToolbarAnimation")
                mt.bind("g", "ToggleGroups")

                mt.bind("v", "ToggleVariableSelection")

                mt.bind("space", "ToggleViewLock", stop_propagation=True)

                mt.bind("esc", "RemoveAllToolbars")

            # Native Dialogs
            client.ClientTriggers(mounted="is_tauri = !!window.__TAURI__")
            with tauri.Dialog() as dialog:
                self.ctrl.save = dialog.save

            with v3.VLayout():
                drawers.Tools(
                    reset_camera=self.view_manager.reset_camera,
                )

                with v3.VMain():
                    dialogs.FileOpen(self.file_browser)
                    dialogs.StateDownload()
                    drawers.FieldSelection(load_variables=self.data_load_variables)

                    with v3.VContainer(classes="h-100 pa-0", fluid=True):
                        with client.SizeObserver("main_size"):
                            # Take space to push content below the fixed overlay
                            html.Div(style=("`height: ${top_padding}px`",))

                            # Fixed overlay for toolbars
                            with html.Div(style=css.TOOLBARS_FIXED_OVERLAY):
                                toolbars.Layout(apply_size=self.view_manager.apply_size)
                                toolbars.Cropping()
                                toolbars.DataSelection()
                                toolbars.Animation()

                            # View of all the variables
                            client.ServerTemplate(
                                name=("active_layout", "auto_layout"),
                                v_if="variables_selected.length",
                            )

                            # Show documentation when no variable selected
                            with html.Div(v_if="!variables_selected.length"):
                                doc.LandingPage()

    # -------------------------------------------------------------------------
    # Derived properties
    # -------------------------------------------------------------------------

    @property
    def selected_variables(self):
        from collections import defaultdict

        vars_per_type = defaultdict(list)
        for var in self.state.variables_selected:
            type = self.source.varmeta[var].dimensions
            vars_per_type[type].append(var)

        return dict(vars_per_type)

    @property
    def selected_variable_names(self):
        # Remove var type (first char)
        return [var for var in self.state.variables_selected]

    # -------------------------------------------------------------------------
    # Methods connected to UI
    # -------------------------------------------------------------------------

    @trigger("download_state")
    @controller.set("download_state")
    def download_state(self):
        active_variables = self.selected_variables
        state_content = {}
        state_content["origin"] = {
            "user": os.environ.get("USER", os.environ.get("USERNAME")),
            "created": f"{datetime.datetime.now()}",
            "comment": self.state.export_comment,
        }
        state_content["files"] = {
            "simulation": str(Path(self.file_browser.get("data_simulation")).resolve()),
            "connectivity": str(
                Path(self.file_browser.get("data_connectivity")).resolve()
            ),
        }
        state_content["variables-selection"] = self.state.variables_selected
        state_content["layout"] = {
            "aspect-ratio": self.state.aspect_ratio,
            "grouped": self.state.layout_grouped,
            "active": self.state.active_layout,
            "tools": self.state.active_tools,
            "help": not self.state.compact_drawer,
        }
        state_content["data-selection"] = {
            k: self.state[k]
            for k in [
                "time_idx",
                "midpoint_idx",
                "interface_idx",
                "crop_longitude",
                "crop_latitude",
                "projection",
            ]
        }
        views_to_export = state_content["views"] = []
        for view_type, var_names in active_variables.items():
            for var_name in var_names:
                config = self.view_manager.get_view(var_name, view_type).config
                views_to_export.append(
                    {
                        "type": view_type,
                        "name": var_name,
                        "config": {
                            # lut
                            "preset": config.preset,
                            "invert": config.invert,
                            "use_log_scale": config.use_log_scale,
                            # layout
                            "order": config.order,
                            "size": config.size,
                            "offset": config.offset,
                            "break_row": config.break_row,
                            # color range
                            "override_range": config.override_range,
                            "color_range": config.color_range,
                            "color_value_min": config.color_value_min,
                            "color_value_max": config.color_value_max,
                        },
                    }
                )

        return json.dumps(state_content, indent=2)

    @change("upload_state_file")
    def _on_import_state(self, upload_state_file, **_):
        if upload_state_file is None:
            return

        file_proxy = file_upload.ClientFile(upload_state_file)
        state_content = json.loads(file_proxy.content)
        self.import_state(state_content)

    @controller.set("import_state")
    def import_state(self, state_content):
        asynchronous.create_task(self._import_state(state_content))

    async def _import_state(self, state_content):
        # Files
        self.file_browser.set_data_simulation(state_content["files"]["simulation"])
        self.file_browser.set_data_connectivity(state_content["files"]["connectivity"])
        await self.data_loading_open(
            self.file_browser.get("data_simulation_files"),
            self.file_browser.get("data_connectivity"),
        )

        # Load variables
        self.state.variables_selected = state_content["variables-selection"]
        self.state.update(state_content["data-selection"])
        await self._data_load_variables()
        self.state.variables_loaded = True

        # Update view states
        for view_state in state_content["views"]:
            view_type = view_state["type"]
            var_name = view_state["name"]
            config = self.view_manager.get_view(var_name, view_type).config
            config.update(**view_state["config"])

        # Update layout
        self.state.aspect_ratio = state_content["layout"]["aspect-ratio"]
        self.state.layout_grouped = state_content["layout"]["grouped"]
        self.state.active_layout = state_content["layout"]["active"]
        self.state.active_tools = state_content["layout"]["tools"]
        self.state.compact_drawer = not state_content["layout"]["help"]

        # Update filebrowser state
        with self.state:
            self.file_browser.set("state_loading", False)

    @controller.add_task("file_selection_load")
    async def data_loading_open(self, simulation_files, connectivity):
        # Reset state
        self.state.variables_selected = []
        self.state.variables_loaded = False
        self.state.midpoint_idx = 0
        self.state.midpoints = []
        self.state.interface_idx = 0
        self.state.interfaces = []
        self.state.time_idx = 0
        self.state.timestamps = []

        self.state.data_files = simulation_files
        self.state.control_data = ""
        self.state.test_data = []

        # Initialize ctrl/test file selection
        if len(simulation_files) >= 2:
            self.state.ctrl_simulation_file = simulation_files[0]
            self.state.test_simulation_file = simulation_files[1]
        elif len(simulation_files) == 1:
            self.state.ctrl_simulation_file = simulation_files[0]
            self.state.test_simulation_file = simulation_files[0]

        await asyncio.sleep(0.1)
        self.source.Update(
            ctrl_file=simulation_files[0],
            test_file=simulation_files[1] if len(simulation_files) > 1 else simulation_files[0],
            conn_file=connectivity,
        )

        self.file_browser.loading_completed(self.source.valid)

        if self.source.valid:
            with self.state as s:
                s.active_tools = list(
                    set(
                        (
                            "select-fields",
                            *(tool for tool in s.active_tools if tool != "load-data"),
                        )
                    )
                )

                self.state.variables_filter = ""
                self.state.variables_listing = [
                    *(
                        {
                            "name": var.name,
                            "type": ", ".join(var.dimensions),
                            "id": f"{var.name}",
                        }
                        for _, var in self.source.varmeta.items()
                    ),
                ]

                # Build dynamic type-color mapping
                from e3sm_quickview.utils.colors import get_type_color

                dim_types = sorted(
                    set(
                        ", ".join(var.dimensions)
                        for var in self.source.varmeta.values()
                    )
                )
                self.state.variable_types = [
                    {"name": t, "color": get_type_color(i)}
                    for i, t in enumerate(dim_types)
                ]

                # Update Layer/Time values and ui layout
                n_cols = 0
                available_tracks = []
                for name, dim in self.source.dimmeta.items():
                    values = dim.data
                    # Convert to list for JSON serialization
                    self.state[name] = (
                        values.tolist()
                        if hasattr(values, "tolist")
                        else list(values)
                        if values is not None
                        else []
                    )
                    if values is not None and len(values) > 1:
                        n_cols += 1
                        available_tracks.append({"title": name, "value": name})
                self.state.toolbar_slider_cols = 12 / n_cols if n_cols else 12
                self.state.animation_tracks = available_tracks
                self.state.animation_track = (
                    self.state.animation_tracks[0]["value"]
                    if available_tracks
                    else None
                )

                from functools import partial

                # Initialize dynamic index variables for each dimension
                for track in available_tracks:
                    dim_name = track["value"]
                    index_var = f"{dim_name}_idx"
                    if "time" in index_var:
                        self.state[index_var] = 50
                    else:
                        self.state[index_var] = 0
                    self.state.change(index_var)(
                        partial(self._on_slicing_change, dim_name, index_var)
                    )

    @controller.set("file_selection_cancel")
    def data_loading_hide(self):
        self.state.active_tools = [
            tool for tool in self.state.active_tools if tool != "load-data"
        ]

    def data_load_variables(self):
        asynchronous.create_task(self._data_load_variables())

    async def _data_load_variables(self):
        """Called at 'Load Variables' button click"""
        vars_to_show = self.selected_variables
        print("Loading variables:", vars_to_show)

        # Flatten the list of lists
        flattened_vars = [var for var_list in vars_to_show.values() for var in var_list]

        self.source.LoadVariables(flattened_vars)

        # Trigger source update + compute avg
        with self.state:
            self.state.variables_loaded = True
        await self.server.network_completion

        await asyncio.sleep(0.1)
        # Use the selected ctrl/test files from the UI state
        ctrl_file = self.state.ctrl_simulation_file or self.source.ctrl_file
        test_file = self.state.test_simulation_file or self.source.test_file

        self.source.Update(
            ctrl_file=ctrl_file,
            test_file=test_file,
            conn_file=self.source.conn_file,
            variables=flattened_vars,
            force_reload=True,
        )

        # Update views in layout
        with self.state:
            self.view_manager.build_auto_layout(vars_to_show)
        await self.server.network_completion

        # Reset camera after yield
        await asyncio.sleep(0.1)
        self.view_manager.reset_camera()

    @change("layout_grouped")
    def _on_layout_change(self, **_):
        vars_to_show = self.selected_variables

        if any(vars_to_show.values()):
            self.view_manager.build_auto_layout(vars_to_show)

    @change("selected_columns")
    def _on_selected_columns_change(self, **_):
        if not self.state.variables_loaded:
            return
        self.view_manager.update_color_range()
        self.view_manager.render()

    @change("projection")
    async def _on_projection(self, projection, **_):
        proj_str = projection[0]
        self.source.UpdateProjection(proj_str)
        self.source.UpdatePipeline()
        self.view_manager.reset_camera()

        # Hack to force reset_camera for "cyl mode"
        # => may not be needed if we switch to rca
        if " " in proj_str:
            for _ in range(2):
                await asyncio.sleep(0.1)
                self.view_manager.reset_camera()

    @change("active_tools", "animation_tracks")
    def _on_toolbar_change(self, active_tools, **_):
        top_padding = 0
        for name in active_tools:
            if name == "select-slice-time":
                track_count = len(self.state.animation_tracks or [])
                rows_needed = max([1, (track_count + 2) // 3])  # 3 sliders per row
                top_padding += 70 * rows_needed
            else:
                top_padding += toolbars.SIZES.get(name, 0)

        self.state.top_padding = top_padding

    def _on_slicing_change(self, var, ind_var, **_):
        self.source.UpdateSlicing(var, self.state[ind_var])
        self.source.UpdatePipeline()

        self.view_manager.update_color_range()
        self.view_manager.render()

        # Update avg computation
        # Get area variable to calculate weighted average
        data = self.source.views["atmosphere_data"]
        self.state.fields_avgs = compute.extract_avgs(
            data, self.selected_variable_names
        )

    @change(
        # "variables_loaded",
        "crop_longitude",
        "crop_latitude",
        "projection",
    )
    def _on_downstream_change(
        self,
        # variables_loaded,
        crop_longitude,
        crop_latitude,
        projection,
        **_,
    ):
        if not self.state.variables_loaded:
            return

        self.source.ApplyClipping(crop_longitude, crop_latitude)
        self.source.UpdateProjection(projection[0])
        self.source.UpdatePipeline()

        self.view_manager.update_color_range()
        self.view_manager.render()

        # Update avg computation
        # Get area variable to calculate weighted average
        data = self.source.views["atmosphere_data"]
        self.state.fields_avgs = compute.extract_avgs(
            data, self.selected_variable_names
        )

    def toggle_toolbar(self, toolbar_name=None):
        if toolbar_name is None:
            self.state.compact_drawer = True
            self.state.active_tools = []
            return

        if toolbar_name in self.state.active_tools:
            # remove
            self.state.active_tools = [
                n for n in self.state.active_tools if n != toolbar_name
            ]
        else:
            # add
            self.state.active_tools.append(toolbar_name)
            self.state.dirty("active_tools")


# -------------------------------------------------------------------------
# Standalone execution
# -------------------------------------------------------------------------
def main():
    app = EAMApp()
    app.server.start()


if __name__ == "__main__":
    main()
